/*
 *  SSL client demonstration program
 *
 *  Copyright (C) 2006-2015, ARM Limited, All Rights Reserved
 *  SPDX-License-Identifier: Apache-2.0
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); you may
 *  not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  This file is part of mbed TLS (https://tls.mbed.org)
 */

#if !defined(MBEDTLS_CONFIG_FILE)
#include "mbedtls/config.h"
#else
#include MBEDTLS_CONFIG_FILE
#endif

#if defined(MBEDTLS_PLATFORM_C)
#include "mbedtls/platform.h"
#else
#include <stdio.h>
#define mbedtls_fprintf    fprintf
#define mbedtls_printf     printf
#endif



#include "mbedtls/net.h"
#include "mbedtls/debug.h"
#include "mbedtls/ssl.h"
#include "mbedtls/entropy.h"
#include "mbedtls/ctr_drbg.h"
#include "mbedtls/error.h"
#include "mbedtls/certs.h"
#include "mbedtls/x509.h"

#include <string.h>

#define SERVER_PORT "4433"
#define SERVER_NAME "172.16.254.50"
#define SNI "server.example.com"
#define GET_REQUEST "GET / HTTP/1.0\r\n\r\n"
#define CIPHERSUITE "TLS-ECDHE-ECDSA-WITH-AES-128-CCM-8"


#define DEBUG_LEVEL 1

static void my_debug( void *ctx, int level,
                      const char *file, int line,
                      const char *str )
{
    ((void) level);

    mbedtls_fprintf( (FILE *) ctx, "%s:%04d: %s", file, line, str );
    fflush(  (FILE *) ctx  );
}

// ----------------------------------------------------------------------------------
#include "stm32f4xx_hal.h"              // Keil::Device:STM32Cube HAL:Common

RNG_HandleTypeDef hrng;

#ifdef __RTX
extern uint32_t os_time;

uint32_t HAL_GetTick(void) {
  return os_time; 
}
#endif

/* The entropy_poll function is a callback provided to 
   mbedtls_entropy_add_source().
	 
	 The arguments are: 
	 - 'data' contains a pointer to the RNG_HandleTypeDef 
	   structure
   - 'output' is a buffer of length 'len' allocated by 
	   the calling function
   - 'olen' is a return value indicating how many 
	   random bytes have been placed in the 'output' 
		 buffer. 
	  
*/
int entropy_poll( void *data, unsigned char *output,
                             size_t len, size_t *olen )
{
	RNG_HandleTypeDef *pRNG=(RNG_HandleTypeDef *) data; 
	uint32_t i=0; 
	uint32_t o_rand, n_rand,init=1; 
	HAL_StatusTypeDef ret; 
	
	if (data == NULL) { 
		mbedtls_printf("\n    No pointer to RNG_HandleTypeDef provided in entropy_poll function call ! " );				
		return (MBEDTLS_ERR_ENTROPY_SOURCE_FAILED );
	}		

  memset(output,0x0,len);
	
	/* Only fetch a whole number of words. The STM32F407 RNG produces 
	   random numbers in 32 bit length. 
	*/
	len = (len /4 ) * 4; 

	while (i < len) { 

		/* In the case of a seed error, the generation of random numbers is 
		   interrupted for as long as the SECS bit is '1'. 
		*/
		
		if (__HAL_RNG_GET_FLAG(pRNG, RNG_FLAG_SECS)==SET) {

						mbedtls_printf("\n    HAL_ERROR - RNG_FLAG_SECS (seed error)! " );
						return (HAL_ERROR );
		}

		/* In the case of a clock error, the RNG is no more able to generate random 
		   numbers because the RNG_CLK clock is not correct.
		*/
		
		if (__HAL_RNG_GET_FLAG(pRNG, RNG_FLAG_CECS)==SET) {

						mbedtls_printf("\n    HAL_ERROR - RNG_FLAG_CECS (clock error)! " );
						return (HAL_ERROR );
		}
		
		/* Section 24.4.3 of the reference manual for the 
		   STM32F407 MCU (RM0090) says that the software
		   must check the DRDY bit, i.e. bit position 0 in the 
       RNG_SR register, before reading data from the RNG data 
		   register (RNG_DR). 
			 
			 This check is done in the HAL_RNG_GenerateRandomNumber
			 function.
		*/ 
		ret=HAL_RNG_GenerateRandomNumber(pRNG, &n_rand); 
		
		switch (ret) { 
			case HAL_OK: 
				/* As required by the FIPS PUB 140-2, the first random number 
			     generated after setting the RNGEN bit should not be used, 
				 	 but saved for comparison with the next generated random number. 
				   Each subsequent generated random number has to be compared with 
			     the previously generated number. The test fails if any two 
			     compared numbers are equal (continuous random number generator 
			     test). 
			  */

				if (init==1) { 
					// first random number will be discarded
					init=0; 
				} else {
					// Check whether two consequtive random numbers are equal
					if (o_rand==n_rand) { 
						// error if the two match
						mbedtls_printf("\n    Two random numbers produced the same value ! " );
						return (MBEDTLS_ERR_ENTROPY_SOURCE_FAILED );
					}	else { 
						// everything OK 		
						memcpy(&output[i], &n_rand,4); 
					}
					i+=4; 
				}
				o_rand=n_rand; 
				break; 
			case HAL_ERROR: 
			  mbedtls_printf("\n    HAL_RNG_GenerateRandomNumber() returned HAL_ERROR ! " );
				return (MBEDTLS_ERR_ENTROPY_SOURCE_FAILED );
			default:
			  mbedtls_printf("\n    HAL_RNG_GenerateRandomNumber() returned unknown error ! " );				
				return (MBEDTLS_ERR_ENTROPY_SOURCE_FAILED );
		}
	}

	*olen = i;

	return( 0 );
}

/// System Clock Configuration
void SystemClock_Config(void) {
  RCC_OscInitTypeDef RCC_OscInitStruct;
  RCC_ClkInitTypeDef RCC_ClkInitStruct;

  /* Enable Power Control clock */
  __HAL_RCC_PWR_CLK_ENABLE();

  /* The voltage scaling allows optimizing the power consumption when the
     device is clocked below the maximum system frequency (see datasheet). */
  __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE1);

  /* Enable HSE Oscillator and activate PLL with HSE as source */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSE;
  RCC_OscInitStruct.HSEState = RCC_HSE_ON;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
  RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSE;
  RCC_OscInitStruct.PLL.PLLM = 25;
  RCC_OscInitStruct.PLL.PLLN = 336;
  RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV2;
  RCC_OscInitStruct.PLL.PLLQ = 7;
  HAL_RCC_OscConfig(&RCC_OscInitStruct);

  /* Select PLL as system clock source and configure the HCLK, PCLK1 and PCLK2
     clocks dividers */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_PCLK1 |
                                RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV4;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV2;
  HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_5);
}
// ----------------------------------------------------------------------------------


/* RNG init function */
void MX_RNG_Init(void)
{

  hrng.Instance = RNG;
  if (HAL_RNG_Init(&hrng) != HAL_OK)
  {
    mbedtls_printf( " HAL_RNG_Init failed\n");
  }

}


int main( void )
{
    int ret, len;
    mbedtls_net_context server_fd;
    unsigned char buf[1024];
    const char *pers = "ssl_client1-2d89f879-0887-4916-8664-30a8c76eaeca"; // device-unique personalization string
	  int force_ciphersuite[2];
	
 const char ca_cert_ec[]=
"-----BEGIN CERTIFICATE-----\r\n"   \
"MIIBYTCCAQegAwIBAgIJAP+YEE1vLE9eMAoGCCqGSM49BAMCMA0xCzAJBgNVBAMM\r\n"   \
"AkNBMB4XDTE4MDMwNzExNTYxMloXDTI4MDMwNDExNTYxMlowDTELMAkGA1UEAwwC\r\n"   \
"Q0EwWTATBgcqhkjOPQIBBggqhkjOPQMBBwNCAAT2KT69W/hlGksPbzJl4AFq3Eg0\r\n"   \
"CFcvxv4vK+vZFYVBmnzjpROm0p3Al0T8FjjxwpFwHTb/3X+txKTZi7oR0WXto1Aw\r\n"   \
"TjAdBgNVHQ4EFgQUJJ+BCK0NP+4gwy+ahsSI2RbXSA0wHwYDVR0jBBgwFoAUJJ+B\r\n"   \
"CK0NP+4gwy+ahsSI2RbXSA0wDAYDVR0TBAUwAwEB/zAKBggqhkjOPQQDAgNIADBF\r\n"   \
"AiEAvmLA5mss65qOy7Hr6wP+vPZUruAlwIbkDE5oY/dDP7ACIANpT8Q/CWN3+YUR\r\n"   \
"lTFQGrT/w9iZuXCVwKbbuFgEAv+W\r\n"   \
"-----END CERTIFICATE-----\r\n"; 
 
const size_t ca_cert_ec_len = sizeof( ca_cert_ec );

const char client_cert_ec[] =
"-----BEGIN CERTIFICATE-----\r\n"   \
"MIIBHTCBxQIJAP43UasTc/ppMAoGCCqGSM49BAMCMA0xCzAJBgNVBAMMAkNBMB4X\r\n"   \
"DTE4MDMwNzEyNDEzN1oXDTI4MDMwNDEyNDEzN1owIjEgMB4GA1UEAwwXQUMtREUt\r\n"   \
"NDgtMjMtNDUtNjctQUItQ0QwWTATBgcqhkjOPQIBBggqhkjOPQMBBwNCAAQIHsdH\r\n"   \
"5vU/EnaOU4y6LOgPnnRjBvRQZ2gRnr6R1ZEXHe4g3HAkBeumW9nFXOtD1F9Olnrw\r\n"   \
"7y9iht3MR+RgUHxbMAoGCCqGSM49BAMCA0cAMEQCIBJjQRL8P9kAO7LUm50uwTAp\r\n"   \
"Ilw4fOFk/CH32lM8rwLPAiABn+h6yoB6MpUjgCnv9oNNIL2M0DaOVi/QU5uDPInE\r\n"   \
"2A==\r\n"   \
"-----END CERTIFICATE-----\r\n";
const size_t client_cert_ec_len = sizeof( client_cert_ec );

const char client_key_ec[] =
"-----BEGIN EC PRIVATE KEY-----\r\n"   \
"MHcCAQEEIFgJd9OYHGpgiUuZhf9SaqcLSTXtY8KZAMIh6F+RzWp3oAoGCCqGSM49\r\n"   \
"AwEHoUQDQgAECB7HR+b1PxJ2jlOMuizoD550Ywb0UGdoEZ6+kdWRFx3uINxwJAXr\r\n"   \
"plvZxVzrQ9RfTpZ68O8vYobdzEfkYFB8Ww==\r\n"   \
"-----END EC PRIVATE KEY-----\r\n";

const size_t client_key_ec_len = sizeof( client_key_ec );

    mbedtls_x509_crt cacert;
    mbedtls_x509_crt clicert;
    mbedtls_pk_context pkey;
    uint32_t flags;
    mbedtls_ecp_group_id curve_list[] = {MBEDTLS_ECP_DP_SECP256R1,MBEDTLS_ECP_DP_NONE};
	
    mbedtls_entropy_context entropy;
    mbedtls_ctr_drbg_context ctr_drbg;
    mbedtls_ssl_context ssl;
    mbedtls_ssl_config conf;

    SystemClock_Config ();

#if defined(MBEDTLS_DEBUG_C)
    mbedtls_debug_set_threshold( DEBUG_LEVEL );
#endif

    /*
     * Initialize TLS session data
     */
    mbedtls_net_init( &server_fd );
    mbedtls_ssl_init( &ssl );
    mbedtls_ssl_config_init( &conf );
		
    mbedtls_x509_crt_init( &cacert );
    mbedtls_x509_crt_init( &clicert );
    mbedtls_pk_init( &pkey );

    /*
     * Initialize the RNG
     */
    mbedtls_ctr_drbg_init( &ctr_drbg );
    MX_RNG_Init();

    mbedtls_printf( "\n  . Seeding the random number generator..." );
    fflush( stdout );

    mbedtls_entropy_init( &entropy );
		
		if (( ret = mbedtls_entropy_add_source(&entropy,entropy_poll,(void*)&hrng,MBEDTLS_CTR_DRBG_ENTROPY_LEN,MBEDTLS_ENTROPY_SOURCE_STRONG))!=0)
		{
		    mbedtls_printf( " failed\n  ! mbedtls_entropy_add_source returned %d\n", ret );
        goto exit;
    }

    mbedtls_ctr_drbg_set_prediction_resistance(&ctr_drbg,MBEDTLS_CTR_DRBG_PR_ON);
		
    if( ( ret = mbedtls_ctr_drbg_seed( &ctr_drbg, mbedtls_entropy_func, &entropy,
                               (const unsigned char *) pers,
                               strlen( pers ) ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_ctr_drbg_seed returned %d\n", ret );
        goto exit;
    }

    mbedtls_printf( " ok\n" );
		
		/*
     * Parse trust anchor / CA cert
     */
    mbedtls_printf( "  . Parse trust anchor ..." );
    fflush( stdout );
		
		ret = mbedtls_x509_crt_parse( &cacert,
                                  (const unsigned char *) ca_cert_ec,
                                  ca_cert_ec_len );

    if( ret != 0 ) { 
        mbedtls_printf( " failed\n  ! mbedtls_x509_crt_parse returned %d\n", ret );
        goto exit;
		}
		
    mbedtls_printf( " ok\n" );
		
    /*
     * Parse own certificate and private key
     * (can be skipped if client authentication is not required)
     */
    mbedtls_printf( "  . Loading the client cert and private key..." );
    fflush( stdout );

    ret = mbedtls_x509_crt_parse( &clicert, (const unsigned char *) client_cert_ec, client_cert_ec_len );

    if( ret != 0 )
    {
        mbedtls_printf( " failed\n  !  mbedtls_x509_crt_parse returned -0x%x\n\n", -ret );
        goto exit;
    }
		
		ret = mbedtls_pk_parse_key( &pkey, (const unsigned char *) client_key_ec, client_key_ec_len, NULL, 0 );
  	
		if( ret != 0 )
    {
        mbedtls_printf( " failed\n  !  mbedtls_pk_parse_key returned -0x%x\n\n", -ret );
        goto exit;
    }
		
    mbedtls_printf( " ok\n" );
		
    /*
     * Establish TCP connection
     */
    mbedtls_printf( "  . Connecting to tcp/%s/%s...", SERVER_NAME, SERVER_PORT );
    fflush( stdout );

    if( ( ret = mbedtls_net_connect( &server_fd, SERVER_NAME,
                                         SERVER_PORT, MBEDTLS_NET_PROTO_TCP ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_net_connect returned %d\n\n", ret );
        goto exit;
    }

    mbedtls_printf( " ok\n" );

    /*
     * Configure TLS
     */
    mbedtls_printf( "  . Setting up the SSL/TLS structure..." );
    fflush( stdout );

    if( ( ret = mbedtls_ssl_config_defaults( &conf,
                    MBEDTLS_SSL_IS_CLIENT,
                    MBEDTLS_SSL_TRANSPORT_STREAM,
                    MBEDTLS_SSL_PRESET_DEFAULT ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_ssl_config_defaults returned %d\n\n", ret );
        goto exit;
    }

    mbedtls_printf( " ok\n" );

		// Configure RNG for use with TLS
    mbedtls_ssl_conf_rng( &conf, mbedtls_ctr_drbg_random, &ctr_drbg );
    // Configure debugging support for TLS
    mbedtls_ssl_conf_dbg( &conf, my_debug, stdout );

		// Configure CA cert
    mbedtls_ssl_conf_ca_chain( &conf, &cacert, NULL );
		
		// Configure client cert / private key
    if( ( ret = mbedtls_ssl_conf_own_cert( &conf, &clicert, &pkey ) ) != 0 )
    {
				mbedtls_printf( " failed\n  ! mbedtls_ssl_conf_own_cert returned %d\n\n", ret );
				goto exit;
    }	
		
    if( ( ret = mbedtls_ssl_setup( &ssl, &conf ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_ssl_setup returned %d\n\n", ret );
        goto exit;
    }
		
		// configure curve
		mbedtls_ssl_conf_curves( &conf, curve_list );
		
		// Configure use of specific ciphersuite
		force_ciphersuite[0] = mbedtls_ssl_get_ciphersuite_id( CIPHERSUITE );

    if( force_ciphersuite[0] != 0 )
    {
				mbedtls_ssl_conf_ciphersuites( &conf, force_ciphersuite );
    }
					
		// Configure send/receive functions
    mbedtls_ssl_set_bio( &ssl, &server_fd, mbedtls_net_send, mbedtls_net_recv, NULL );

		if( ( ret = mbedtls_ssl_set_hostname( &ssl, SNI ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_ssl_set_hostname returned %d\n\n", ret );
        goto exit;
    }
		
    /*
     * Run TLS handshake protocol
     */
    mbedtls_printf( "  . Performing the TLS handshake..." );
    fflush( stdout );

    while( ( ret = mbedtls_ssl_handshake( &ssl ) ) != 0 )
    {
        if( ret != MBEDTLS_ERR_SSL_WANT_READ && ret != MBEDTLS_ERR_SSL_WANT_WRITE )
        {
            mbedtls_printf( " failed\n  ! mbedtls_ssl_handshake returned -0x%x\n\n", -ret );
					 if( ret == MBEDTLS_ERR_X509_CERT_VERIFY_FAILED )
                mbedtls_printf(
                    "    Unable to verify the server's certificate. "
                        "Either it is invalid,\n"
                    "    or you didn't set ca_file or ca_path "
                        "to an appropriate value.\n"
                    "    Alternatively, you may want to use "
                        "auth_mode=optional for testing purposes.\n" );
            mbedtls_printf( "\n" );
            goto exit;
        }
    }

    mbedtls_printf( " ok\n" );

		/*
     * Verify the server certificate
     */
    mbedtls_printf( "  . Verifying server certificate..." );

    if( ( flags = mbedtls_ssl_get_verify_result( &ssl ) ) != 0 )
    {
        char vrfy_buf[512];

        mbedtls_printf( " failed\n" );

        mbedtls_x509_crt_verify_info( vrfy_buf, sizeof( vrfy_buf ), "  ! ", flags );

        mbedtls_printf( "%s\n", vrfy_buf );
    }
    else
        mbedtls_printf( " ok\n" );

    if( mbedtls_ssl_get_peer_cert( &ssl ) != NULL )
    {
        mbedtls_printf( "  . Peer certificate information    ...\n" );
        mbedtls_x509_crt_info( (char *) buf, sizeof( buf ) - 1, "      ",
                       mbedtls_ssl_get_peer_cert( &ssl ) );
        mbedtls_printf( "%s\n", buf );
    }
		
    /*
     * Write the GET request
     */
    mbedtls_printf( "  > Write to server:" );
    fflush( stdout );

    len = sprintf( (char *) buf, GET_REQUEST );

    while( ( ret = mbedtls_ssl_write( &ssl, buf, len ) ) <= 0 )
    {
        if( ret != MBEDTLS_ERR_SSL_WANT_READ && ret != MBEDTLS_ERR_SSL_WANT_WRITE )
        {
            mbedtls_printf( " failed\n  ! mbedtls_ssl_write returned %d\n\n", ret );
            goto exit;
        }
    }

    len = ret;
    mbedtls_printf( " %d bytes written\n\n%s", len, (char *) buf );

    /*
     * Read the HTTP response
     */
    mbedtls_printf( "  < Read from server:" );
    fflush( stdout );

    do
    {
        len = sizeof( buf ) - 1;
        memset( buf, 0, sizeof( buf ) );
        ret = mbedtls_ssl_read( &ssl, buf, len );

        if( ret == MBEDTLS_ERR_SSL_WANT_READ || ret == MBEDTLS_ERR_SSL_WANT_WRITE )
            continue;

        if( ret == MBEDTLS_ERR_SSL_PEER_CLOSE_NOTIFY )
            break;

        if( ret < 0 )
        {
            mbedtls_printf( "failed\n  ! mbedtls_ssl_read returned %d\n\n", ret );
            break;
        }

        if( ret == 0 )
        {
            mbedtls_printf( "\n\nEOF\n\n" );
            break;
        }

        len = ret;
        mbedtls_printf( " %d bytes read\n\n%s", len, (char *) buf );
    }
    while( 1 );

    mbedtls_ssl_close_notify( &ssl );

	  /*
     * Cleanup and exit
     */
		
exit:

#ifdef MBEDTLS_ERROR_C
    if( ret != 0 )
    {
        char error_buf[100];
        mbedtls_strerror( ret, error_buf, 100 );
        mbedtls_printf("Last error was: %d - %s\n\n", ret, error_buf );
    }
#endif

    mbedtls_net_free( &server_fd );

		mbedtls_x509_crt_free( &clicert );
    mbedtls_x509_crt_free( &cacert );
    mbedtls_pk_free( &pkey );
		
    mbedtls_ssl_free( &ssl );
    mbedtls_ssl_config_free( &conf );
    mbedtls_ctr_drbg_free( &ctr_drbg );
    mbedtls_entropy_free( &entropy );

		// Shell can not handle large exit numbers -> 1 for errors
    if( ret < 0 )
        ret = 1;
		
    return( ret );
}

